import os
import numpy as np
import iris
import iris.cube
import iris.analysis
import warnings
import numba
import datetime
from numba import jit
from numba import vectorize
from numba import guvectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver

MESH_FILE = "/esarchive/autosubmit/con_files/mesh_mask_nemo.Ec3.2_O1L75.nc"
DATA_FILE = "/esarchive/exp/ecearth/t05l/cmorfiles/CMIP/EC-Earth-Consortium" \
            "/EC-Earth3-LR/historical/r1i1p1f1/Omon/thetao/gn/v20190205" \
            "/thetao_Omon_EC-Earth3-LR_historical_r1i1p1f1_gn_199001-199001.nc"
REGIONS_FILE = "/esarchive/autosubmit/con_files/mask.regions.Ec3.2_O1L75.nc"


def main():
    start = datetime.datetime.now()
    one_layer = ((0, 300),)
    # multiple_layers = ((0, 200), (200, 700), (700, 2000),
    #                    (0, 100), (100, 200),(200, 300), (300, 400),
    #                    (400, 500), (500, 1000),)

    mask, e3t, depth = load_mesh()
    weights = check_device_and_compute_weights(one_layer, mask, e3t, depth)
    del mask, depth, e3t
    basin_name, basins = load_basins()
    e1t, e2t = load_areas()
    area = {}
    for basin in basins:
        area[basin] = compute_area_basin(e1t, e2t, basins[basin])
    del e1t, e2t, basins
    thetao = load_thetao()
    thetao += 273.15
    ohc, ohc_1D = compute_OHC(one_layer, weights, thetao, area)
    del weights, thetao, area
    if numba.cuda.is_available() is False:
        device = 'CPU'
    else:
        device = 'GPU'
    save_data(one_layer, basin_name, ohc, ohc_1D, device)
    seconds_total = datetime.datetime.now() - start
    print('Total ellapsed time on the {1}: {0}'.format(seconds_total, device))


def check_device_and_compute_weights(layers, mask, e3t, depth):
    check_gpu = numba.cuda.is_available()
    if check_gpu is False:
        device = 'CPU'
        weights = compute_weights_cpu(layers, mask, e3t, depth)
    else:
        use_gpu = True
        if use_gpu is True:
            device = 'GPU'
            weights = compute_weights_gpu(layers, mask, e3t, depth)
        else:
            device = 'CPU'
            weights = compute_weights_cpu(layers, mask, e3t, depth)
    print('Running on {}'.format(device))
    return weights


def compute_OHC(layers, weights, thetao, area):
    check_gpu = numba.cuda.is_available()
    if check_gpu is False:
        ohc, ohc_1D = compute_ohc_cpu(layers, thetao, weights, area)
    else:
        ohc, ohc_1D = compute_ohc_gpu(layers, thetao, weights, area)
    return ohc, ohc_1D


def load_mesh():
    mask = iris.load_cube(MESH_FILE, 'tmask').data.astype(np.float32)
    e3t = iris.load_cube(MESH_FILE, 'e3t_0').data.astype(np.float32)
    depth = iris.load_cube(MESH_FILE, 'gdepw_0').data.astype(np.float32)
    return mask, e3t, depth


def compute_weights_cpu(layers, mask, e3t, depth):
    weights = []
    for min_depth, max_depth in layers:
        weights.append(calculate_weight_numba(min_depth, max_depth, e3t, depth,
                       mask))
    return weights


def compute_weights_gpu(layers, mask, e3t, depth):
    gpu_mask = cuda.to_device(mask.data.astype(np.float32))
    gpu_e3t = cuda.to_device(e3t.data.astype(np.float32))
    gpu_depth = cuda.to_device(depth.data.astype(np.float32))
    weights = []
    for min_depth, max_depth in layers:
        weights.append(calculate_weight_numba_cuda(min_depth, max_depth,
                       gpu_e3t, gpu_depth, gpu_mask))
    del gpu_depth, gpu_mask, gpu_e3t
    return weights


def load_basins():
    regions = iris.load(REGIONS_FILE)
    basins = {}
    basin_name = []
    for region in regions:
        name = region.name()
        if name in ('nav_lat', 'nav_lon'):
            continue
        basin_name.append(name)
        extract_region = region.extract(iris.Constraint(z=1))
        basins[name] = extract_region.data.astype(np.float32)
    return basin_name, basins


def load_areas():
    e1t = iris.load_cube(MESH_FILE, 'e1t').data.astype(np.float32)
    e2t = iris.load_cube(MESH_FILE, 'e2t').data.astype(np.float32)
    return e1t, e2t


def load_thetao():
    thetao = iris.load_cube(DATA_FILE, 'sea_water_potential_temperature')
    thetao_data = thetao.data.astype(np.float32)
    return thetao_data


def compute_ohc_cpu(layers, thetao, weights, area):
    ohc = []
    for layer in range(len(layers)):
        ohc_layer = sum_red.reduce(multiply_thetao_weight(thetao,
                                   weights[layer]), axis=1)
        ohc.append(ohc_layer)
        ohc1D_total = []
        for basin in area:
            ohc_basin = multiply_array(ohc_layer, area[basin])
            temp = sum_red.reduce(ohc_basin, axis=1)
            ohc1D = sum_red.reduce(temp, axis=1)
            ohc1D_total.append(ohc1D)
    return ohc, ohc1D_total


def compute_ohc_gpu(layers, thetao, weights, area):
    levels = thetao.shape[1]
    times = thetao.shape[0]
    lats = thetao.shape[2]
    lons = thetao.shape[3]
    basins = len(area)

    area_basin = np.empty((basins, lats, lons))
    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)
    ohc = []
    gpu_ohc = cuda.device_array((times, lats, lons), dtype=np.float32)
    gpu_temp = cuda.device_array((basins, times, lats*lons), dtype=np.float32)
    for i, basin in enumerate(area):
        area_basin[i, :, :] = area[basin]
    gpu_basins_area = cuda.to_device(area_basin.astype(np.float32))
    gpu_thetao = cuda.to_device(thetao.astype(np.float32))

    for layer in range(len(layers)):
        compute_ohc[grid, block](gpu_thetao, weights[layer], gpu_ohc, levels)
        ohc.append(gpu_ohc.copy_to_host())  # moure al final
        multiply_ohc_basin[grid, block](gpu_ohc, gpu_basins_area, gpu_temp)
        ohc1D_basin = []
        for basin in range(basins):
            ohc_1D = np.empty(times, dtype=np.float32)
            for time in range(times):
                ohc_1D[time] = sum_red_cuda(gpu_temp[basin, time, :])
            ohc1D_basin.append(ohc_1D)
    del gpu_ohc, gpu_temp, gpu_basins_area

    return ohc, ohc1D_basin


def save_data(layers, basins, ohc, ohc1D, device):
    ohc_cube = []
    print(len(ohc1D))
    for i, layer in enumerate(layers):
        ohc_cube.append(iris.cube.Cube(ohc[i],
                                       long_name='Ocean heat content'
                                                 ' {0[0]} to {0[1]} meters'
                                       .format(layer)))
        ohc_1D = []
        for j, basin in enumerate(basins):
            ohc_1D.append(iris.cube.Cube(ohc1D[j][:],
                          long_name='{0}'.format(basin)))
        iris.save(ohc_1D,
                  '/esarchive/scratch/sloosvel/numba_outputs'
                  '/ohc_1D_{1}_{0[0]}_{0[1]}_numba_vec.nc'
                  .format(layer, device), zlib=True)
    iris.save(ohc_cube,
              '/esarchive/scratch/sloosvel/numba_outputs/ohc_{0}.nc'
              .format(device), zlib=True)


@vectorize(['float32(int32, int32, float32, float32, float32)'], target='cuda')
def calculate_weight_numba_cuda(min_depth, max_depth, e3t, depth, mask):
    """
    Calculates the weight for each cell
    """
    if not mask:
        return 0
    top = depth
    bottom = top + e3t
    if bottom < min_depth or top > max_depth:
        return 0
    else:
        if top < min_depth:
            top = min_depth
        if bottom > max_depth:
            bottom = max_depth

        return (bottom - top) * 1020 * 4000


@cuda.jit()
def compute_ohc(thetao, weight, ohc, levels):
    i = cuda.threadIdx.x + cuda.blockIdx.x * cuda.blockDim.x
    j = cuda.blockIdx.y
    t = cuda.blockIdx.z
    # i, j, t = grid(3)
    ohc[t, j, i] = 0.0
    temp = 0.0
    if(i >= thetao.shape[3]):
        return
    for lev in range(levels):
        thetao_value = thetao[t, lev, j, i]
        temp += thetao_value * weight[0, lev, j, i]
    ohc[t, j, i] = temp


@cuda.jit()
def multiply_ohc_basin(ohc, area, temp):
    i, j, t = cuda.grid(3)
    basins = area.shape[0]
    x = i+j*ohc.shape[2]
    if(i >= ohc.shape[2]):
        return
    for basin in range(basins):
        temp[basin, t, x] = 0.0
        temp[basin, t, x] = ohc[t, j, i] * area[basin, j, i]


@vectorize(['float32(int32, int32, float32, float32, float32)'], target='cpu')
def calculate_weight_numba(min_depth, max_depth, e3t, depth, mask):
    """
    Calculates the weight for each cell
    """
    if not mask:
        return 0
    top = depth
    bottom = top + e3t
    if bottom < min_depth or top > max_depth:
        return 0
    else:
        if top < min_depth:
            top = min_depth
        if bottom > max_depth:
            bottom = max_depth

        return (bottom - top) * 1020 * 4000


@vectorize(['float32(float32, float32)'], target='cpu')
def multiply_thetao_weight(thetao, weight):
    return thetao*weight


@vectorize(['float64(float32, float32)', 'float64(float64, float64)'],
           target='cpu')
def sum_red(x, y):
    return x+y


@vectorize(['float32(float32, float32)', 'float64(float64, float64)'],
           target='cpu')
def multiply_array(a, b):
    return a*b


@vectorize(['float32(float32, float32, float32)'], target='cpu')
def compute_area_basin(e1t, e2t, basin):
    return e1t*e2t*basin


@cuda.reduce
def sum_red_cuda(x, y):
    return x+y


if __name__ == '__main__':
    main()
